/*---------------------------------------------------------------------------*\

    FILE....: HIP.CPP
    TYPE....: C+ Functions
    AUTHOR..: David Rowe
    DATE....: 19/11/97

    VPB access functions that implement the PC side of the Host Interface 
    Port (HIP) - the interface between the PC and DSP.

\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\

         Voicetronix Voice Processing Board (VPB) Software

         Copyright (C) 1999-2001 Voicetronix www.voicetronix.com.au

         This library is free software; you can redistribute it and/or
         modify it under the terms of the GNU Lesser General Public
         License as published by the Free Software Foundation; either
         version 2.1 of the License, or (at your option) any later version.

         This library is distributed in the hope that it will be useful,
         but WITHOUT ANY WARRANTY; without even the implied warranty of
         MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
         Lesser General Public License for more details.

         You should have received a copy of the GNU Lesser General Public
         License along with this library; if not, write to the Free Software
         Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
	 USA

\*---------------------------------------------------------------------------*/

#include "hip.h"
#include "port.h"
#include "vpbioctl.h"
#include "generic.h"

#include <assert.h>
#include <stdio.h>
#include <memory.h>

/*-------------------------------------------------------------------------*\

				DEFINES

\*-------------------------------------------------------------------------*/

/* general */

#define	SPAN		4	/* number of I/O ports used 	*/
#define	NO_VPB		0xffff	/* indicates no valid VPB	*/

/* HIP registers */

#define	DATA		0x0	/* offset of data register	*/
#define	CNTRL		0x1	/* offset of control register	*/
#define	ADDRLO		0x2	/* low byte of address register	*/
#define	ADDRHI		0x3	/* hi byte of address register	*/

/* Control register bits */

#define	RESET		0x1	/* assert to place DSP in reset	*/
#define	CNTEN		0x2	/* assert to enable addr cntr	*/
#define UP		0x4     /* ==1 for up count, else down	*/

/*-------------------------------------------------------------------------*\

				CLASSES

\*-------------------------------------------------------------------------*/

// pure virtual base class implementation

class baseHipData {
public:	
	baseHipData() {}
	virtual ~baseHipData() {}
	virtual void InitVpb(USHORT base) = 0;
	virtual void DspReset(USHORT board) = 0;
	virtual void DspRun(USHORT board) = 0;
	virtual void WriteDspSram(USHORT board, USHORT addr, USHORT length, 
				  word *buf) = 0;
	virtual void ReadDspSram(USHORT board, USHORT addr, USHORT length,
				 word *buf) = 0;
#ifndef __FreeBSD__
	virtual void WriteIic(USHORT board, USHORT addr, USHORT length, 
				  word *buf) = 0;
	virtual void ReadIic(USHORT board, USHORT addr, USHORT length,
				 word *buf) = 0;
#endif
	virtual void EeRead(USHORT board, USHORT addr, USHORT length,
				 word *buf) = 0;
	virtual void SetPip(USHORT board, char enables) = 0;
	virtual void *GethndFile() = 0;
	virtual int GetDeviceDriverModel() = 0;
};

// Hip implementation for original device driver

class HipData : public baseHipData {
	static	int exists;		// asserted if HIP module exists	
	USHORT	num_vpb;		// number of VPB cards init-ed		
	USHORT	cntrl[MAX_VPB];	        // mirror of VPB control ports		
	Port	*port;			// lower level port driver
	USHORT	pipmask;		// current pip enable mask
	int		DeviceDriverModel;
	int		os;		// operating system identifier
public:	
	HipData(int DeviceDriverModel);
	~HipData();
	void InitVpb(USHORT base);
	void DspReset(USHORT board);
	void DspRun(USHORT board);
	void WriteDspSram(USHORT board, USHORT addr, USHORT length, word *buf);
	void ReadDspSram(USHORT board, USHORT addr, USHORT length, word *buf);
	void WriteIic(USHORT board, USHORT addr, USHORT length, word *buf);
	void ReadIic(USHORT board, USHORT addr, USHORT length, word *buf);
	void EeRead(USHORT board, USHORT addr, USHORT length, word *buf);
	void SetPip(USHORT board, char enables);
	void *GethndFile() {return 0;}
	int  GetDeviceDriverModel() { return DeviceDriverModel; }
};

int HipData::exists = 0;

// Alternate implementation for V4PCI

class HipDataPCI : public baseHipData {

	int     hndFile;
	int	DeviceDriverModel;
public:	
	HipDataPCI(int DeviceDriverModel);
	~HipDataPCI();
	void InitVpb(USHORT base);
	void DspReset(USHORT board);
	void DspRun(USHORT board);
	void WriteDspSram(USHORT board, USHORT addr, USHORT length, word *buf);
	void ReadDspSram(USHORT board, USHORT addr, USHORT length, word *buf);
	void WriteIic(USHORT board, USHORT addr, USHORT length, word *buf);
	void ReadIic(USHORT board, USHORT addr, USHORT length, word *buf);
	void EeRead(USHORT board, USHORT addr, USHORT length, word *buf);
	void SetPip(USHORT board, char enables);
	void *GethndFile() {return (void*)hndFile;}
	int  GetDeviceDriverModel() { return DeviceDriverModel; }
};

// used to make access to DSP thread safe

static GENERIC_CRITICAL_SECTION	HipSect;

/*-------------------------------------------------------------------------*\

			     HIP MEMBER FUNCTIONS

\*-------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*\

    FUNCTION.: Hip::Hip
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    Initialises Hip states and lower level device driver software.

\*-------------------------------------------------------------------------*/

Hip::Hip(int DeviceDriverModel)
{
	if (DeviceDriverModel == 0) {
		// original driver

		d = new HipData(DeviceDriverModel);
		if (d == NULL)
			assert(0); // HIP_CANT_ALLOCATE_MEMORY
	}

	if (DeviceDriverModel == 1) {
		assert(0); // not supported in thsi version of driver
	}

	if (DeviceDriverModel == 2) {
		// PCI driver (Linux only)

		d = new HipDataPCI(DeviceDriverModel);
		if (d == NULL)
			assert(0); // HIP_CANT_ALLOCATE_MEMORY
	}
}

/*-------------------------------------------------------------------------*\

    FUNCTION.: Hip::~Hip()
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    Closes down the HIP software and lower level drivers.

\*-------------------------------------------------------------------------*/

Hip::~Hip()
{
	delete d;
}

/*-------------------------------------------------------------------------*\

    FUNCTION.: Hip:InitVpb
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    Inits a specific VPB card. Must be called before any operations are 
    performed on the VPB.

\*-------------------------------------------------------------------------*/

void Hip::InitVpb(USHORT base)
//      USHORT base;	base address of card
{
	d->InitVpb(base);
}

/*-------------------------------------------------------------------------*\

    FUNCTION.: Hip::DspReset
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    Asserts the reset line of the DSP, freezing code execution on the DSP.

\*-------------------------------------------------------------------------*/

void Hip::DspReset(USHORT board)
//      USHORT board;	VPB board number
{
	d->DspReset(board);
}

/*-------------------------------------------------------------------------*\

    FUNCTION.: Hip::DspRun
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    De-asserts the reset line of the DSP, starting code execution on the DSP.

\*-------------------------------------------------------------------------*/

void Hip::DspRun(USHORT board)
//      USHORT board;	VPB board number
{
	d->DspRun(board);
}

/*-------------------------------------------------------------------------*\
	
    FUNCTION.: Hip::WriteDspSram
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    Writes a block of 16bit words to the DSPs SRAM.  The length and address
    values are both specified in words (not bytes).

\*-------------------------------------------------------------------------*/

void Hip::WriteDspSram(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	start of block in DSP SRAM	
//      USHORT length;	length of block		
//      word   *buf;	ptr to block in PC	
{
	d->WriteDspSram(board, addr, length, buf);
}

/*-------------------------------------------------------------------------*\
	
    FUNCTION.: Hip::ReadDspSram
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    Reads a block of 16bit words from the DSPs SRAM.  The length and address
    values are both specified in words (not bytes).

\*-------------------------------------------------------------------------*/

void Hip::ReadDspSram(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	start of block in DSP SRAM	
//      USHORT length;	length of block		
//      word   *buf;	ptr to block in PC	
{
	d->ReadDspSram(board, addr, length, buf);
}

#ifndef __FreeBSD__
/*-------------------------------------------------------------------------*\
	
    FUNCTION.: Hip::WriteIic
    AUTHOR...: Peter Wintulich
    DATE.....: 


\*-------------------------------------------------------------------------*/

void Hip::WriteIic(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	Upper byte is device address,
//	                lower byte is start of block in IIC device.
//      USHORT length;	length of block		
//      word   *buf;	ptr to block in PC	
{
	d->WriteIic(board, addr, length, buf);
}

/*-------------------------------------------------------------------------*\
	
    FUNCTION.: Hip::ReadIic
    AUTHOR...: Peter Wintulich
    DATE.....: 


\*-------------------------------------------------------------------------*/

void Hip::ReadIic(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	Upper byte is device address,
//		      	lower byte is start of block in IIC device.	
//      USHORT length;	length of block		
//      word   *buf;	ptr to block in PC	
{
	d->ReadIic(board, addr, length, buf);
}

#endif

/*-------------------------------------------------------------------------*\
	
    FUNCTION.: Hip::EeRead
    AUTHOR...: Peter Wintulich
    DATE.....: 


\*-------------------------------------------------------------------------*/

void Hip::EeRead(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	start of block in EeProm	
//      USHORT length;	length of block		
//      word   *buf;	ptr to block in PC	
{
  	d->EeRead(board, addr, length, buf);
}

/*-------------------------------------------------------------------------*\
	
    FUNCTION.: Hip::SetPip
    AUTHOR...: David Rowe
    DATE.....: 9/8/98

    Sets the pip enable mask for a given board.

\*-------------------------------------------------------------------------*/

void Hip::SetPip(USHORT board, char enables)
//      USHORT board;	VPB board number
//      char   enables;	eight bit mask defining which pip channels are enabled
{
	d->SetPip(board, enables);
}

// helper functions

void *Hip::GethndFile() {return d->GethndFile();}
int  Hip::GetDeviceDriverModel() { return d->GetDeviceDriverModel(); }

/*-------------------------------------------------------------------------*\

			   HIPDATA MEMBER FUNCTIONS

\*-------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*\

    FUNCTION.: HipData::HipData
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    Initialises Hip states and lower level device driver software.

\*-------------------------------------------------------------------------*/

HipData::HipData(int Ddm)
{
	int    i;
	
	DeviceDriverModel = Ddm;
	if (exists) 
		assert(0); // HIP_ALREADY_INITIALISED
	exists = 1;

	GenericInitializeCriticalSection(&HipSect);

	for(i=0; i<MAX_VPB; i++)
		cntrl[i] = 0;
	num_vpb = 0;
	
	// determine operating system

	os = GenericGetOS();
	if (os == GET_OS_FAIL)
		assert(0); // HIP_CANT_GET_OS_VER
	
	// initialise appropriate port class for this OS

	switch(os) {
	case GET_OS_LINUX:
		// Linux specific port I/O
		port = new PortLinux;
		if (port == NULL)
			assert(0); // HIP_CANT_ALLOCATE_MEMORY
		break;
	case GET_OS_USER:
		assert(0);
		#ifdef NOT_USED
		// Generic port I/O with user mode port access
		port = new PortUser;
		if (port == NULL)
			assert(0); // HIP_CANT_ALLOCATE_MEMORY
		break;
		#endif
	default:
		assert(0);         // HIP_OS_NOT_SUPPORTED
	}

	pipmask = 0;
}

/*-------------------------------------------------------------------------*\

    FUNCTION.: HipData::~HipData()
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    Closes down the HIP software and lower level drivers.

\*-------------------------------------------------------------------------*/

HipData::~HipData()
{
	int	i;

	/* Place all DSPs in RESET */

	GenericEnterCriticalSection(&HipSect);
	for(i=0; i<num_vpb; i++) {
		cntrl[i] = RESET;
		port->write(i, CNTRL, cntrl[i]);
	}

	// delete appropriate port class for this OS

	switch(os) {
	case GET_OS_LINUX:
		delete (PortLinux*)port;
		break;
	case GET_OS_USER:
		delete (PortUser*)port;
		break;
	default:
		assert(0); // HIP_OS_NOT_SUPPORTED
	}

	exists = 0;
	GenericLeaveCriticalSection(&HipSect);

	GenericDeleteCriticalSection(&HipSect);
}

/*-------------------------------------------------------------------------*\

    FUNCTION.: HipData:InitVpb
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    Inits a specific VPB card. Must be called before any operations are 
    performed on the VPB.

\*-------------------------------------------------------------------------*/

void HipData::InitVpb(USHORT base)
//      USHORT base;	base address of card
{
	// check range of base address

	assert((base >= START_BASE) && (base < END_BASE));

	// now tell port about new board and reset board

	GenericEnterCriticalSection(&HipSect);
	port->addBoard(base, SPAN);
	cntrl[num_vpb] = 0;
	port->write(num_vpb, CNTRL, cntrl[num_vpb]);
	num_vpb++;
	GenericLeaveCriticalSection(&HipSect);
}

/*-------------------------------------------------------------------------*\

    FUNCTION.: HipData::DspReset
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    Asserts the reset line of the DSP, freezing code execution on the DSP.

\*-------------------------------------------------------------------------*/

void HipData::DspReset(USHORT board)
//      USHORT board;	VPB board number
{
	assert(board < num_vpb);
	GenericEnterCriticalSection(&HipSect);
	cntrl[board] |= RESET;
	port->write(board, CNTRL, cntrl[board]);
	GenericLeaveCriticalSection(&HipSect);
}

/*-------------------------------------------------------------------------*\

    FUNCTION.: HipData::DspRun
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    De-asserts the reset line of the DSP, starting code execution on the DSP.

\*-------------------------------------------------------------------------*/

void HipData::DspRun(USHORT board)
//      USHORT board;	VPB board number
{
 	assert(board < num_vpb);
	GenericEnterCriticalSection(&HipSect);
	cntrl[board] &= ~RESET;
	port->write(board, CNTRL, cntrl[board]);
	GenericLeaveCriticalSection(&HipSect);
}

/*-------------------------------------------------------------------------*\
	
    FUNCTION.: HipData::WriteDspSram
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    Writes a block of 16bit words to the DSPs SRAM.  The length and address
    values are both specified in words (not bytes).

\*-------------------------------------------------------------------------*/

void HipData::WriteDspSram(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	start of block in DSP SRAM	
//      USHORT length;	length of block		
//      word   *buf;	ptr to block in PC	
{
	/* validate arguments */

	assert(board < num_vpb);
//	assert(addr < MAX_LENGTH);
//	assert(length < MAX_LENGTH);
	assert(buf != NULL);

	/* set up for transfer */

	GenericEnterCriticalSection(&HipSect);
	cntrl[board] |= CNTEN;
	cntrl[board] |= UP;
	port->write(board, CNTRL, cntrl[board]);
	port->write(board, ADDRLO, pipmask + (addr & 0xff));
	port->write(board, ADDRHI, (addr>>8) & 0xff);

	/* perform transfer */

	port->blockWrite(board, DATA, length, buf);
	GenericLeaveCriticalSection(&HipSect);
}

/*-------------------------------------------------------------------------*\
	
    FUNCTION.: Hip::ReadDspSram
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    Reads a block of 16bit words from the DSPs SRAM.  The length and address
    values are both specified in words (not bytes).

\*-------------------------------------------------------------------------*/

void HipData::ReadDspSram(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	start of block in DSP SRAM	
//      USHORT length;	length of block		
//      word   *buf;	ptr to block in PC	
{
        /* validate arguments */

	assert(board < num_vpb);
//	assert(addr < MAX_LENGTH);
//	assert(length < MAX_LENGTH);
	assert(buf != NULL);

	/* set up for transfer */

	GenericEnterCriticalSection(&HipSect);
	cntrl[board] |= CNTEN;
	cntrl[board] |= UP;
	port->write(board, CNTRL, cntrl[board]);
	port->write(board, ADDRLO, pipmask + (addr & 0xff));
	port->write(board, ADDRHI, (addr>>8) & 0xff);

	/* perform transfer */

	port->blockRead(board, DATA, length, buf);
	GenericLeaveCriticalSection(&HipSect);
}

/*-------------------------------------------------------------------------*\
	
    FUNCTION.: Hip::ReadIic
    AUTHOR...: 
    DATE.....: 

    Dummy function as ISA dose not use PLX interface chip

\*-------------------------------------------------------------------------*/

void HipData::ReadIic(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	IIc Dev + offset	
//      USHORT length;	length of block		
//      word   *buf;	ptr to block in PC	
	//      Not supported under ISA only for PCI PLX based cards
{
        int i;
	
	/* validate arguments */

	assert(board < num_vpb);
//	assert(addr < MAX_LENGTH);
//	assert(length < MAX_LENGTH);
	assert(buf != NULL);

	/* set up for transfer */

	GenericEnterCriticalSection(&HipSect);
	// ISA CARD no support fill with zeros
	for(i=0; i<length; i++)
	       	*(buf+i)= 0;
	GenericLeaveCriticalSection(&HipSect);
}


/*-------------------------------------------------------------------------*\
	
    FUNCTION.: Hip::WriteIic
    AUTHOR...: 
    DATE.....: 

    Dummy function as ISA dose not use PLX interface chip

\*-------------------------------------------------------------------------*/

void HipData::WriteIic(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	Dev & offset	
//      USHORT length;	length of block		
//      word   *buf;	ptr to block in PC	
	//      Not supported under ISA only for PCI PLX based cards
{
        int i;
	
	/* validate arguments */

	assert(board < num_vpb);
//	assert(addr < MAX_LENGTH);
//	assert(length < MAX_LENGTH);
	assert(buf != NULL);

	/* set up for transfer */

	GenericEnterCriticalSection(&HipSect);
	// ISA CARD no support fill with zeros
	for(i=0; i<length; i++)
	       	*(buf+i)= 0;
	GenericLeaveCriticalSection(&HipSect);
}


/*-------------------------------------------------------------------------*\
	
    FUNCTION.: Hip::EeRead
    AUTHOR...: David Rowe
    DATE.....: 19/11/97

    Dummy function as ISA dose not use PLX interface chip

\*-------------------------------------------------------------------------*/

void HipData::EeRead(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	start of block in DSP SRAM	
//      USHORT length;	length of block		
//      word   *buf;	ptr to block in PC	
	//      Not supported under ISA only for PCI PLX based cards
{
        int i;
	
	/* validate arguments */

	assert(board < num_vpb);
//	assert(addr < MAX_LENGTH);
//	assert(length < MAX_LENGTH);
	assert(buf != NULL);

	/* set up for transfer */

	GenericEnterCriticalSection(&HipSect);
	// ISA CARD no support fill with zeros
	for(i=0; i<length; i++)
	       	*(buf+i)= 0;
	GenericLeaveCriticalSection(&HipSect);
}

/*-------------------------------------------------------------------------*\
	
    FUNCTION.: HipData::SetPip
    AUTHOR...: David Rowe
    DATE.....: 9/8/98

    Sets the pip enable mask for a given board.

\*-------------------------------------------------------------------------*/

void HipData::SetPip(USHORT board, char enables)
//  USHORT board;	VPB board number
//  USHORT enables;	eight bit mask defining which pip channels are enabled
{
    assert(board < num_vpb);
    pipmask = enables<<8;
}

/*-------------------------------------------------------------------------*\

			  HIPDATAPCI MEMBER FUNCTIONS

\*-------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*\

    FUNCTION.: HipDataPCI::HipDataPCI
    AUTHOR...: David Rowe
    DATE.....: 2/10/98

    Initialises Hip states and lower level device driver software.

\*-------------------------------------------------------------------------*/

HipDataPCI::HipDataPCI(int Ddm)
{
	DeviceDriverModel = Ddm;

	// check this operating system supported

	int os = GenericGetOS();
	if (os == -1)
		assert(0); // HIP_CANT_GET_OS_VER
	
	// initialise appropriate port class for this OS

	switch(os) {
		case GET_OS_LINUX:
		break;
		default:
			assert(0); // HIP_OS_NOT_SUPPORTED
	}

	// Open device driver

	hndFile = Generic_open();

        if (hndFile == -1)        // Was the device opened?
		assert(0); // HIP_CANT_START_DEVICE_DRIVER

	// Tell DD to open hip module
	GenericInitializeCriticalSection(&HipSect);
}

/*-------------------------------------------------------------------------*\

    FUNCTION.: HipDataPCI::~HipDataPCI()
    AUTHOR...: David Rowe
    DATE.....: 2/10/98

    Closes down the HIP software and lower level drivers.

\*-------------------------------------------------------------------------*/

HipDataPCI::~HipDataPCI()
{
	DspReset(0);
	Generic_close(hndFile);
	GenericDeleteCriticalSection(&HipSect);
}

/*-------------------------------------------------------------------------*\

    FUNCTION.: HipDataPCI:InitVpb
    AUTHOR...: David Rowe
    DATE.....: 2/10/98

    Inits a specific VPB card. Must be called before any operations are 
    performed on the VPB.

\*-------------------------------------------------------------------------*/

void HipDataPCI::InitVpb(USHORT base)
//      USHORT base;	base address of card
{
}

/*-------------------------------------------------------------------------*\

    FUNCTION.: HipDataPCI::DspReset
    AUTHOR...: David Rowe
    DATE.....: 2/10/98

    Asserts the reset line of the DSP, freezing code execution on the DSP.

\*-------------------------------------------------------------------------*/

void HipDataPCI::DspReset(USHORT board)
//  USHORT board;	VPB board number
{
	int ret;

	GenericEnterCriticalSection(&HipSect);

	ret = Generic_pci_dsp_reset(hndFile, board);
	if (ret == -1)                          
		assert(0); // HIP_DSP_RESET_FAIL

	GenericLeaveCriticalSection(&HipSect);
}

/*-------------------------------------------------------------------------*\

    FUNCTION.: HipDataPCI::DspRun
    AUTHOR...: David Rowe
    DATE.....: 2/10/98

    De-asserts the reset line of the DSP, starting code execution on the DSP.

\*-------------------------------------------------------------------------*/

void HipDataPCI::DspRun(USHORT board)
//      USHORT board;	VPB board number
{
	int ret;

	GenericEnterCriticalSection(&HipSect);

	ret = Generic_pci_dsp_run(hndFile, board);
	if (ret == -1)                          
		assert(0); // HIP_DSP_RUN_FAIL

	GenericLeaveCriticalSection(&HipSect);
}

/*-------------------------------------------------------------------------*\
	
    FUNCTION.: HipDataPCI::WriteDspSram
    AUTHOR...: David Rowe
    DATE.....: 2/10/98

    Writes a block of 16bit words to the DSPs SRAM.  The length and address
    values are both specified in words (not bytes).

\*-------------------------------------------------------------------------*/

void HipDataPCI::WriteDspSram(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	start of block in DSP SRAM	
//      USHORT length;	length of block		
//      word   *buf;	ptr to block in PC	
{
	// validate arguments

//	assert(addr < MAX_LENGTH);
//	assert(length < MAX_LENGTH);
	assert(buf != NULL);

	int ret;

	GenericEnterCriticalSection(&HipSect);

	ret = Generic_pci_block_write(hndFile, board, addr, length, buf);
	if (ret == -1)                          
		assert(0); // HIP_WRITE_DSP_SRAM_FAIL
 
	GenericLeaveCriticalSection(&HipSect);
}

/*-------------------------------------------------------------------------*\
	
    FUNCTION.: HipDataPCI::ReadDspSram
    AUTHOR...: David Rowe
    DATE.....: 2/10/98

    Reads a block of 16bit words from the DSPs SRAM.  The length and address
    values are both specified in words (not bytes).

\*-------------------------------------------------------------------------*/

void HipDataPCI::ReadDspSram(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	start of block in DSP SRAM	
//      USHORT length;	length of block		
//      word   *buf;	ptr to block in PC	
{
	// validate arguments

//	assert(addr < MAX_LENGTH);
//	assert(length < MAX_LENGTH);
	assert(buf != NULL);

	int ret;

	GenericEnterCriticalSection(&HipSect);

	ret = Generic_pci_block_read(hndFile, board, addr, length, buf);
	if (ret == -1)                          
		assert(0); // HIP_READ_DSP_SRAM_FAIL

	GenericLeaveCriticalSection(&HipSect);
}

#ifndef __FreeBSD__
/*-------------------------------------------------------------------------*\
	
    FUNCTION.: HipDataPCI::WriteIic
    AUTHOR...: 
    DATE.....: 

    Writes a block of 16bit 

\*-------------------------------------------------------------------------*/

void HipDataPCI::WriteIic(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	start of block in DSP SRAM	
//      USHORT length;	length of block		
//      word   *buf;	ptr to block in PC	
{
	// validate arguments

//	assert(addr < MAX_LENGTH);
//	assert(length < MAX_LENGTH);
	assert(buf != NULL);

	int ret;

	GenericEnterCriticalSection(&HipSect);

	ret = Generic_pci_block_iicwrite(hndFile, board, addr, length, buf);
	if (ret == -1)                          
		assert(0); // HIP_WRITE_I2C_FAIL
 
	GenericLeaveCriticalSection(&HipSect);
}

/*-------------------------------------------------------------------------*\
	
    FUNCTION.: HipDataPCI::ReadIic
    AUTHOR...: 
    DATE.....: 

    Reads a block of 16bit 

\*-------------------------------------------------------------------------*/

void HipDataPCI::ReadIic(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	start of block in DSP SRAM	
//      USHORT length;	length of block		
//      word   *buf;	ptr to block in PC	
{
	// validate arguments

//	assert(addr < MAX_LENGTH);
//	assert(length < MAX_LENGTH);
	assert(buf != NULL);

	int ret;

	GenericEnterCriticalSection(&HipSect);

	ret = Generic_pci_block_iicread(hndFile, board, addr, length, buf);
	if (ret == -1)                          
		assert(0); // HIP_READ_I2C_FAIL

	GenericLeaveCriticalSection(&HipSect);
}
#endif

/*-------------------------------------------------------------------------*\
	
    FUNCTION.: HipDataPCI::EeRead
    AUTHOR...: Peter Wintulich
    DATE.....: 12/02/2003

    Reads a block of 16bit words from the PLX Config EEROM.  The length and
    address values are both specified in words (not bytes).

\*-------------------------------------------------------------------------*/

void HipDataPCI::EeRead(USHORT board, USHORT addr, USHORT length, word *buf)
//      USHORT board;	VPB board number
//      USHORT addr;	start location of EEROM	read
//      USHORT length;	Number of words to read		
//      word   *buf;	ptr to block in PC	
{
	// validate arguments

//	assert(addr < MAX_LENGTH);
//	assert(length < MAX_LENGTH);
	assert(buf != NULL);

	int ret;

	GenericEnterCriticalSection(&HipSect);

	ret = Generic_pci_block_eeread(hndFile, board, addr, length, buf);
	if (ret == -1)                          
		assert(0); // HIP_READ_EEROM_FAIL

	GenericLeaveCriticalSection(&HipSect);
}


/*-------------------------------------------------------------------------*\
	
    FUNCTION.: HipDataPCI::SetPip
    AUTHOR...: David Rowe
    DATE.....: 9/8/98

    Sets the pip enable mask for a given (VPB8L) board.  Does nothing as
    no PCI VPB8L exists (yet!).

\*-------------------------------------------------------------------------*/

void HipDataPCI::SetPip(USHORT board, char enables)
//      USHORT board;	VPB board number
//      USHORT enables;	eight bit mask defining which pip channels are enabled
{
}



